import argparse
from .fetchers import FetcherEnum


def get_parser():
	"""
	Function used to parse the arguments in the command line interface

	:return: object containing the parse results
	"""

	parser = argparse.ArgumentParser()
	subparsers = parser.add_subparsers(dest="subparser")
	autodl = subparsers.add_parser("autodl", help="auto downloader using the mangas in the json")
	manga = subparsers.add_parser("manga", help="tool to modify, add and remove mangas from the automatic rss downloader mode list")
	manual_pars = subparsers.add_parser("manual", help="manually download scans, it will not update the downloaded scans json, if you plan on setting up a manga with the automatic rss mode don't mix both commands")

	parser.add_argument("-q", "--quiet", action="store_true", help="removes the verbose")
	parser.add_argument("--version", action="store_true", help="print the current software version")


	# manual downloading subparser
	sauce = manual_pars.add_mutually_exclusive_group(required=True)
	sauce.add_argument("-l", "--link", type=str, help="gives the link to the page with all the chapter listed")
	sauce.add_argument("-m", "--manga", type=str, help="gives the manga name, the image fetcher will find the corresponding manga")

	manual_pars.add_argument("-f", "--fetcher", type=str, required=True, help="the name of the image links fetcher that will be used for the download")
	manual_pars.add_argument("-o", "--output", default=".", type=str, help="the path (absolute or relative) to the folder where to save the data, the images will be stored in a subfolder images with inside one folder per chapter and the pdfs will be stored in a pdf subfolder")
	manual_pars.add_argument("-c", "--chapter-start", default=1, type=str, help="gives the chapter to start the download on (defaults at 1)")

	amount = manual_pars.add_mutually_exclusive_group()
	amount.add_argument("-a", "--all", action="store_true", help="downloads all the chapters from the starting point to the end")
	amount.add_argument("-n", "--download-number", default=1, type=int, help="gives the number of chapters to download (defaults at 1)")
	amount.add_argument("-e", "--chapter-end", default=0, type=int, help="gives the ending chapter number, if the chapter doesn't exists the program will stop once it's surpassed")

	manual_pars.add_argument("-t", "--tiny", action="store_true", help="don't write the manga name in the title (useful if using ebook libraries)")
	manual_pars.add_argument("-s", "--skip", default=0, type=int, help="skips n images before starting to download the first chapter")

	content = manual_pars.add_mutually_exclusive_group(required=True)
	content.add_argument("-p", "--pdf", action="store_true", help="downloads only the pdf of the manga")
	content.add_argument("-k", "--keep", action="store_true", help="downloads the pdf but also keep the images in a chapter subfolder")
	content.add_argument("-i", "--image", action="store_true", help="downloads only the images in a chapter subfolder and don't create the pdf")


	# json interaction subparser
	list_type = manga.add_mutually_exclusive_group()
	list_type.add_argument("-l", "--list", action="store_true", help="list all the non-archived mangas for autodl")
	list_type.add_argument("-la", "--list-all", action="store_true", help="list all the mangas for autodl")
	list_type.add_argument("-lo", "--list-only", action="store_true", help="list only the archived mangas for autodl")

	db_management = manga.add_mutually_exclusive_group()
	db_management.add_argument("-e", "--export-db", type=str, help="exports the current database into a json file")
	db_management.add_argument("-i", "--import-db", type=str, help="imports a new database from a json file")
	manga_subparser = manga.add_subparsers(dest="manga_subparser")

	## add subparser
	add = manga_subparser.add_parser("add", help="add a new manga to the auto downloader")
	add.add_argument("name", type=str, help="name for the stored manga")
	add.add_argument("-l", "--link", type=str, help="link of the manga", required=True)
	add.add_argument("-f", "--fetcher", type=str, help="name of the fetcher needed for the manga", required=True)
	add.add_argument("-c", "--chap", type=str, nargs="*", help="list of all the chapters already downloaded to be added to the list for the auto-updater")
	add.add_argument("-a", "--archived", action="store_true", help="create the manga as archived")

	## edit subparser
	edit = manga_subparser.add_parser("edit", help="modify infos for one of the already existing manga in the auto downloader")
	edit.add_argument("name", type=str, help="name for the stored manga")
	edit.add_argument("-l", "--link", type=str, help="link of the manga")
	edit.add_argument("-f", "--fetcher", type=str, help="name of the fetcher needed for the manga")
	edit.add_argument("-c", "--chap", type=str, nargs="*", help="list of all the chapters already downloaded to be added to the list for the auto-updater")

	archiving = edit.add_mutually_exclusive_group()
	archiving.add_argument("-a", "--archive", action="store_true", help="makes the edited manga archived")
	archiving.add_argument("-u", "--unarchive", action="store_true", help="unarchives the edited manga")


	## info subparser
	info = manga_subparser.add_parser("info", help="prints the info for the named manga in the auto updater")
	info.add_argument("name", type=str, help="name for the stored manga")

	## chapter list subparser
	chaplist = manga_subparser.add_parser("chaplist", help="lists all the dowloaded chapters (warning: can be huge)")
	chaplist.add_argument("name", type=str, help="name for the stored manga")

	## delete subparser
	delete = manga_subparser.add_parser("delete", help="deletes the corresponding manga from the auto downloader")
	delete.add_argument("name", type=str, help="name for the stored manga")

	## remove chapters subparser
	rmchaps = manga_subparser.add_parser("rmchaps", help="remove the listed chapters for this manga from the database")
	rmchaps.add_argument("name", type=str, help="name for the stored manga")
	rmchaps.add_argument("chap", type=str, nargs="*", help="remove the listed chapters for this manga from the database")


	# autodl subparser
	autodl.add_argument("-o", "--output", type=str, default=".", help="the path (absolute or relative) to the folder where to save the data")
	autodl.add_argument("-t", "--tiny", action="store_true", help="don't write the manga name in the title (useful if using ebook libraries)")
	content = autodl.add_mutually_exclusive_group(required=True)
	content.add_argument("-p", "--pdf", action="store_true", help="downloads only the pdf of the manga")
	content.add_argument("-k", "--keep", action="store_true", help="downloads the pdf but also keep the images in a chapter subfolder")
	content.add_argument("-i", "--image", action="store_true", help="downloads only the images in a chapter subfolder and don't create the pdf")


	manual_pars.epilog = "The current list of image fetcher is: " + ", ".join(FetcherEnum.list())
	manga.epilog = manual_pars.epilog

	return parser



from .excepts import DryNoSauceHere, TooManySauce, EmptyChapter, DelayedRelease
from .fetchers.fetcher import StandaloneFetcher
from PIL import Image
import img2pdf
import requests
import logging
import os
import io
from re import sub as re_sub


class Pyscandl:
	"""
	The main object of the program. It is responsible of the downloads and controls the fetchers for you.
	"""
	def __init__(self, fetcher, chapstart=1, output:str=".", pdf:bool=True, keep:bool=False, image:bool=False, all:bool=False, link:str=None, manga:str=None, download_number:int=1, chapend=0, quiet:bool=False, skip:int=0, tiny:bool=False):
		"""
		Initialize this instance of the pyscandl downloader, it needs either manga or link to work.

		:param fetcher: fetcher object related to the download
		:param chapstart: first chapter to be downloaded
		:type chapstart: int/float/str
		:param output: output folder
		:type output: str
		:param pdf: tell if the result should be kept as a pdf
		:type pdf: bool
		:param keep: tell if the result should be kept as a pdf and as a collection of images
		:type keep: bool
		:param image: tell if the result should be kept as a collection of images
		:type image: bool
		:param all: download all the chapters that are available after chapstart
		:type all: bool
		:param link: link of the manga to download
		:type link: str
		:param manga: identification tag of the manga *(see every fetcher for their different variations)*
		:type manga: str
		:param download_number: number of chapters to download
		:type download_number: int
		:param chapend: chapter to end the download on, if non exstant the download will stop once the next to download chapter number is greater than it
		:type chapend: int/float/str
		:param quiet: should the program not output any information about what it is doing in the console
		:type quiet: bool
		:param skip: number of images to skip on the first chapter being downloaded *(useful if running in image mode)*
		:type skip: int
		:param tiny: should the name of every downloaded scan be minified and only include the chapter number and the chapter title
		:type tiny: bool

		:raises DryNoSauceHere: neither link or manga was specified
		:raises TooManySauce: both link and manga were specified
		"""

		if link is not None and manga is None or link is None and manga is not None:
			if issubclass(fetcher, StandaloneFetcher):
				self.fetcher = fetcher(link=link, manga=manga)
			else:
				self.fetcher = fetcher(link=link, manga=manga, chapstart=chapstart)
		elif link is None and manga is None:
			raise DryNoSauceHere
		else:
			raise TooManySauce

		# in case windows is the os, remove the banned characters
		if os.name == "nt":
			manga_name = re_sub(r'[\\/*?:"<>|]', u"█", self.fetcher.manga_name)
		else:
			manga_name = self.fetcher.manga_name

		# creating output folder
		self._output = (output[-1] == "/" and output or output + "/") + manga_name + "/"

		if not os.path.exists(self._output):
			os.makedirs(self._output)

		self._header = {"User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/77.0.3865.120 Safari/537.36",
					   "Set-Cookie": f"domain={self.fetcher.domain}"}
		self._header.update(self.fetcher.headers)
		self._nskip = skip
		self._quiet = quiet

		# select download mode
		self._pdf = pdf
		self._keep = keep
		self._image = image

		self._all = all
		self._download_number = download_number
		self._chapend = chapend
		self._path = f"{self._output}ch.{self.fetcher.chapter_number} {self.fetcher.chapter_name}/"  # save path for images
		self._img_bin_list = []
		self._tiny = tiny

		# in case windows is the os, remove the banned characters
		if os.name == "nt":
			chapter_name = re_sub(r'[\\/*?:"<>|]', u"█", self.fetcher.chapter_name)
		else:
			chapter_name = self.fetcher.chapter_name

		if self._tiny:
			if isinstance(self.fetcher, StandaloneFetcher):
				self._pdf_path = f"{self._output}{chapter_name}.pdf"
				self._name_metadata_pdf = f"{self.fetcher.chapter_name}"
			else:
				self._pdf_path = f"{self._output}ch.{self.fetcher.chapter_number} {chapter_name}.pdf"
				self._name_metadata_pdf = f"ch.{self.fetcher.chapter_number} {self.fetcher.chapter_name}"
		else:
			if isinstance(self.fetcher, StandaloneFetcher):
				self._pdf_path = f"{self._output}{manga_name} - {chapter_name}.pdf"
				self._name_metadata_pdf = f"{self.fetcher.manga_name} - {self.fetcher.chapter_name}"
			else:
				self._pdf_path = f"{self._output}{manga_name} - ch.{self.fetcher.chapter_number} {chapter_name}.pdf"
				self._name_metadata_pdf = f"{self.fetcher.manga_name} - ch.{self.fetcher.chapter_number} {self.fetcher.chapter_name}"

		self._banlist = []
		ban_path = f"{os.path.dirname(os.path.abspath(__file__))}/../banlist"
		for img in os.listdir(ban_path):
			with open(f"{ban_path}/{img}", "rb") as img_bin:
				self._banlist.append(img_bin.read())

		# disabling logging for alpha-channels for img2pdf
		logging.disable()

	def _dl_image(self):
		"""
		Downloads the currently selected image.
		"""

		if not os.path.exists(self._path):
			os.makedirs(self._path)

		with open(f"{self._path}{self.fetcher.npage}.{self.fetcher.ext}", "wb") as img:
			img.write(requests.get(self.fetcher.image, headers=self._header).content)
			if not self._quiet:
				print(".", end="", flush=True)

	def full_chapter(self):
		"""
		Fetching all the images of the chapter and storing them in RAM.
		"""

		if not self._quiet:
			if isinstance(self.fetcher, StandaloneFetcher):
				print(f"fetching: {self.fetcher.chapter_name}")
			else:
				print(f"fetching: ch.{self.fetcher.chapter_number} {self.fetcher.chapter_name}")
		while not self.fetcher.is_last_image():
			self._img_bin_list.append(requests.get(self.fetcher.image, headers=self._header).content)
			if not self._quiet:
				print(".", end="", flush=True)
			self.fetcher.next_image()
		self._img_bin_list.append(requests.get(self.fetcher.image, headers=self._header).content)
		if not self._quiet:
			print(".", end="", flush=True)

	def keep_full_chapter(self):
		"""
		Downloading all the images of the chapters and storing them where the output was specified.
		"""

		if not self._quiet:
			if isinstance(self.fetcher, StandaloneFetcher):
				print(f"downloading: {self.fetcher.chapter_name}")
			else:
				print(f"downloading: ch.{self.fetcher.chapter_number} {self.fetcher.chapter_name}")
		while not self.fetcher.is_last_image():
			self._dl_image()
			self.fetcher.next_image()
		self._dl_image()
		if not self._quiet and self._image:
			print("")

	def _skip(self):
		"""
		Skips the images as asked with the skip parameter.
		"""

		for loop in range(self._nskip):
			self.fetcher.next_image()

	def create_pdf(self):
		"""
		Creates the pdf at the output location with the fetched or the downloaded images of the current chapter.

		:raises EmptyChapter: the images of the current chapter were all blacklisted images and the pdf was empty
		"""
		if not self._quiet:
			print("\nconverting...", end=" ")
		# loading the downloaded images if keep mode
		if self._keep:
			for loop in range(1, self.fetcher.npage + 1):
				try:
					with open(f"{self._path}{loop}.jpg", "rb") as img:
						self._img_bin_list.append(img.read())
				except FileNotFoundError:
					try:
						with open(f"{self._path}{loop}.png", "rb") as img:
							self._img_bin_list.append(img.read())
					except FileNotFoundError:
						with open(f"{self._path}{loop}.gif", "rb") as img:
							self._img_bin_list.append(img.read())

		# removing the images found in the banlist
		self._img_bin_list = [img for img in self._img_bin_list if img not in self._banlist]

		if len(self._img_bin_list) > 0:
			# creating the pdf
			try:
				with open(self._pdf_path, "wb") as pdf:
					pdf.write(img2pdf.convert(self._img_bin_list, title=self._name_metadata_pdf, author=self.fetcher.author, keywords=[self.fetcher.manga_name]))
			except Exception as e:
				# removing alpha from all the images of the chapter
				if e.args[0] == "Refusing to work on images with alpha channel":
					if not self._quiet:
						print("removing alpha...", end=" ")

					dealpha_list = []
					for img in self._img_bin_list:
						temp = Image.open(io.BytesIO(img)).convert("RGB")
						with io.BytesIO() as dealpha_img:
							temp.save(dealpha_img, format="JPEG")
							dealpha_list.append(dealpha_img.getvalue())
					with open(self._pdf_path, "wb") as pdf:
						pdf.write(img2pdf.convert(dealpha_list, title=self._name_metadata_pdf, author=self.fetcher.author, keywords=[self.fetcher.manga_name]))
				else:
					raise e

			if not self._quiet:
				print("converted")
		else:
			raise EmptyChapter(self.fetcher.manga_name, self.fetcher.chapter_number)

	def go_to_chapter(self, chap_num):
		"""
		Make Pyscandl go to the asked chapter.

		:param chap_num: chapter number that was asked for
		:type chap_num: int/str/float
		"""

		# in case windows is the os, remove the banned characters
		if os.name == "nt":
			chapter_name = re_sub(r'[\\/*?:"<>|]', u"█", self.fetcher.chapter_name)
		else:
			chapter_name = self.fetcher.chapter_name

		self.fetcher.go_to_chapter(chap_num)
		self._path = f"{self._output}ch.{self.fetcher.chapter_number} {chapter_name}/"
		self._img_bin_list = []
		# prepares the next pdf path and name
		if self._tiny:
			if isinstance(self.fetcher, StandaloneFetcher):
				self._pdf_path = f"{self._output}{chapter_name}.pdf"
				self._name_metadata_pdf = f"{self.fetcher.chapter_name}"
			else:
				self._pdf_path = f"{self._output}ch.{self.fetcher.chapter_number} {chapter_name}.pdf"
				self._name_metadata_pdf = f"ch.{self.fetcher.chapter_number} {self.fetcher.chapter_name}"
		else:
			if isinstance(self.fetcher, StandaloneFetcher):
				self._pdf_path = f"{self._output}{self.fetcher.manga_name} - {chapter_name}.pdf"
				self._name_metadata_pdf = f"{self.fetcher.manga_name} - {self.fetcher.chapter_name}"
			else:
				self._pdf_path = f"{self._output}{self.fetcher.manga_name} - ch.{self.fetcher.chapter_number} {chapter_name}.pdf"
				self._name_metadata_pdf = f"{self.fetcher.manga_name} - ch.{self.fetcher.chapter_number} {self.fetcher.chapter_name}"

	def next_chapter(self):
		"""
		Goes to the next chapter
		"""

		self.fetcher.next_chapter()

		# in case windows is the os, remove the banned characters
		if os.name == "nt":
			chapter_name = re_sub(r'[\\/*?:"<>|]', u"█", self.fetcher.chapter_name)
		else:
			chapter_name = self.fetcher.chapter_name

		self._path = f"{self._output}ch.{self.fetcher.chapter_number} {chapter_name}/"
		self._img_bin_list = []
		# prepares the next pdf path and name
		if self._tiny:
			if isinstance(self.fetcher, StandaloneFetcher):
				self._pdf_path = f"{self._output}{chapter_name}.pdf"
				self._name_metadata_pdf = f"{self.fetcher.chapter_name}"
			else:
				self._pdf_path = f"{self._output}ch.{self.fetcher.chapter_number} {chapter_name}.pdf"
				self._name_metadata_pdf = f"ch.{self.fetcher.chapter_number} {self.fetcher.chapter_name}"
		else:
			if isinstance(self.fetcher, StandaloneFetcher):
				self._pdf_path = f"{self._output}{self.fetcher.manga_name} - {chapter_name}.pdf"
				self._name_metadata_pdf = f"{self.fetcher.manga_name} - {self.fetcher.chapter_name}"
			else:
				self._pdf_path = f"{self._output}{self.fetcher.manga_name} - ch.{self.fetcher.chapter_number} {chapter_name}.pdf"
				self._name_metadata_pdf = f"{self.fetcher.manga_name} - ch.{self.fetcher.chapter_number} {self.fetcher.chapter_name}"

	def full_download(self):
		"""
		Does the full download process with what is specified when initializing the Pyscandl object
		"""

		try:
			# emulating a do while
			self._skip()
			counter = 1

			try:
				if self._keep or self._image:
					self.keep_full_chapter()
				else:
					self.full_chapter()

				if not self._image:
					try:
						self.create_pdf()
					except EmptyChapter:
						if not self._quiet:
							print("empty")
			except DelayedRelease as e:
				if not self._quiet:
					print(e)

			while not isinstance(self.fetcher, StandaloneFetcher) and not self.fetcher.is_last_chapter() and (self._all or counter < self._download_number or float(self.fetcher.chapter_number) < self._chapend):
				self.next_chapter()
				try:
					if self._keep or self._image:
						self.keep_full_chapter()
					else:
						self.full_chapter()

					if not self._image:
						try:
							self.create_pdf()
						except EmptyChapter:
							if not self._quiet:
								print("empty")
				except DelayedRelease as e:
					if not self._quiet:
						print(e)
				counter += 1
		except KeyboardInterrupt:
			if not self._quiet:
				print("\nmanual interruption")
		finally:
			self.fetcher.quit()
			if not self._quiet:
				print("end of the download")



from bs4 import BeautifulSoup
import requests
import re

from pyscandl.modules.fetchers.fetcher import Fetcher
from pyscandl.modules.excepts import MangaNotFound, DryNoSauceHere, DelayedRelease

class FRScan(Fetcher):
    def __init__(self, link: str = None, manga: str = None, chapstart=1):
        # getting the manga id
        super().__init__(chapstart=chapstart)
        if manga is not None:
            self._link = f"https://www.frscan.me/manga/{manga}"
            manga_id = manga
        elif link is not None:
            if link[-1] == "/":
                link = link[:-1]
            manga_id = link.split("/")[-1]
            self._link = link
        else:
            raise DryNoSauceHere()

        req = requests.get(self._link)
        if req.status_code == 404:
            raise MangaNotFound(manga_id)

        manga_soup = BeautifulSoup(req.text, "html.parser")
        chap_tags = manga_soup.find("ul", class_="chapters")

        chap_nb = []
        chap_name = []
        for chap in chap_tags.find_all("a", href=re.compile(rf"{self._link}/\d+(?:\.\d+)?")):
            nb = chap.get("href").split("/")[-1]
            if "." in nb:
                chap_nb.append(float(nb))
            else:
                chap_nb.append(int(nb))

            if chap.find_next_sibling("span").text:
                chap_name.append(-1)
            else:
                chap_name.append(chap.find_next_sibling("em").text)

        self._chaps = dict(zip(chap_nb, chap_name))

        self.author:str = manga_soup.find("dt", string="Auteur(s)").find_next_sibling("dd").findChild("a").text.strip()  #: name of the author of the manga default to TBD if none is found
        self.manga_name:str = manga_soup.find("h2", class_="widget-title").text.strip()  #: name of the manga currentl in the fetcher
        self.domain:str = ".frscan.me"  #: domain of the website the fetcher is currently on looks like: .domname.ext
        self.go_to_chapter(chapstart)

    def go_to_chapter(self, chap):
        self.chapter_number = int(chap)
        self.chapter_name:str = self._chaps.get(self.chapter_number)
        if self.chapter_name is None:
            raise MangaNotFound(f"{self.manga_name}, chapter {self.chapter_number}")
        if self.chapter_name == -1:
            raise DelayedRelease(f"{self.manga_name}, chapter {self.chapter_number}", "later")

        req = requests.get(f"{self._link}/{self.chapter_number}/1")
        self._chap_soup = BeautifulSoup(req.text, "html.parser")

        self.image: str = self._chap_soup.find(class_="img-responsive scan-page").get("src").strip()  #: url to the image currently in the fetcher
        if "http" not in self.image.split("/")[0]:
            self.image = "http:" + self.image
        self.ext: str = self.image.split(".")[-1]  #: extention of the image currently in the fetcher
        self.npage: int = 1  #: number of the page the fetcher is currently on

    def next_image(self):
        self.npage += 1
        req = requests.get(f"{self._link}/{self.chapter_number}/{self.npage}")
        self._chap_soup = BeautifulSoup(req.text, "html.parser")

        self.image: str = self._chap_soup.find(class_="img-responsive scan-page").get("src").strip()  #: url to the image currently in the fetcher
        if "http" not in self.image.split("/")[0]:
            self.image = "http:" + self.image
        self.ext: str = self.image.split(".")[-1]  #: extention of the image currently in the fetcher

    def next_chapter(self):
        for i in sorted(self._chaps):
            if i > self.chapter_number:
                self.go_to_chapter(i)
                break

    def is_last_image(self):
        options = self._chap_soup.find(id="page-list").find_all("option")
        return self.npage == int(max(options, key=lambda option: int(option.get("value"))).get("value"))

    def is_last_chapter(self):
        return self.chapter_number == max(self._chaps.keys())

    @classmethod
    def scan(cls, link: str = None, manga: str = None):
        if manga is not None:
            link = f"https://www.frscan.me/manga/{manga}"
            manga_id = manga
        elif link is not None:
            if link[-1] == "/":
                link = link[:-1]
            manga_id = link.split("/")[-1]
            link = link
        else:
            raise DryNoSauceHere()

        req = requests.get(link)
        if req.status_code == 404:
            raise MangaNotFound(manga_id)

        manga_soup = BeautifulSoup(req.text, "html.parser")
        chap_tags = manga_soup.find("ul", class_="chapters")
        chap_nb = []
        for chap in chap_tags.find_all("a", href=re.compile(rf"{link}/\d+(?:\.\d+)?")):
            nb = chap.get("href").split("/")[-1]
            if "." in nb:
                chap_nb.append(float(nb))
            else:
                chap_nb.append(int(nb))

        return chap_nb


class TooManySauce(Exception):
	__doc__ = """
	Exception used if the user has given a link and a manga name at the same time.
	"""

	def __init__(self):
		Exception.__init__(self, "too many sauce, you can't give a manga and a link at the same time")


class DryNoSauceHere(Exception):
	__doc__ = """
	Exception used if the user has given no link or manga
	"""

	def __init__(self):
		Exception.__init__(self, "you need to give either a link to the manga or its name")


class MangaNotFound(Exception):
	__doc__ = """
	Exception used when the asked manga or chapter can't be found
	"""

	def __init__(self, name):
		"""
		:param name: name of the manga
		:type name: str
		"""
		Exception.__init__(self, f"the manga {name} was not found")


class FetcherNotFound(Exception):
	__doc__ = """
	Exception used when the asked fetcher doesn't exists
	"""

	def __init__(self, fetcher):
		"""
		:param fetcher: name of the fetcher
		:type fetcher: str
		"""
		from .fetchers import fetcher_enum  # future proofed for circular imports
		Exception.__init__(self, f"{fetcher.upper()} is not a supported fetcher the list of supported fetchers is: " + ", ".join(
            fetcher_enum.Fetchers.list()))


class NoFetcherGiven(Exception):
	__doc__ = """
	Exception used when there was no fetcher given to the program
	"""

	def __init__(self):
		Exception.__init__(self, "no fetcher given, please give a fetcher")


class IsStandalone(Exception):
	__doc__ = """
	Exception used when trying to add to autodl a scan using a standalone fetcher
	"""

	def __init__(self, name):
		"""
		:param name: name of the manga using the standalone fetcher
		:type name: str
		"""

		Exception.__init__(self, f"{name} has a standalone type fetcher and thus can't be used with autodl")


class EmptyChapter(Exception):
	__doc__ = """
	Exception used when the current chapter is empty
	"""

	def __init__(self, manga, chap_num):
		"""
		:param manga: name of the manga
		:type manga: str
		:param chap_num: chapter number that is empty in :param manga:
		:type chap_num: int/float/str
		"""

		Exception.__init__(self, f"chapter {chap_num} of {manga} is empty")


class DelayedRelease(Exception):
	__doc__ = "exception used if a chapter is visible on the website but will be available later, this can be the case on mangadex"

	def __init__(self, name, date):
		"""
		:param name: name of the manga
		:param date: date of the availability
		"""

		Exception.__init__(self, f'the manga {name} will be available starting {date}')


class DownedSite(Exception):
	__doc__ = """
	Exception used when the fetcher can't access its website to do the downloads
	"""

	def __init__(self, website_name):
		"""
		:param website_name:
		"""

		Exception.__init__(self, f"The website {website_name} isn't accessible for the moment, please retry it later")



# referencing the fetchers for the PyPI package

from .fanfox import Fanfox

# mangadex setup
from .mangadex import Mangadex
from .mangadex import MangadexEN
from .mangadex import MangadexFR

from .nh import NHentai
from .frscan import FRScan

# naver webtoons
from .naver import NaverWebtoon
from .naver import NaverBestChallenge
from .naver import NaverChallenge

# webtoons setup
from .webtoons import Webtoons
from .webtoons import WebtoonsEN
from .webtoons import WebtoonsFR

# getting the enumeration
from .fetcher_enum import FetcherEnum



from ..excepts import MangaNotFound, EmptyChapter, DelayedRelease, DownedSite
from .fetcher import Fetcher
import cfscrape
import requests
import json.decoder
from datetime import datetime


class Mangadex(Fetcher):
	__doc__ = """
	This is the fetcher in charge of https://mangadex.org/ 
	"""

	def __init__(self, link:str=None, manga:str=None, chapstart=1, lang=""):
		"""
		Initializes the instance of the nhentai fetcher, it needs either manga or link to work.

		:param link: link of the scan wanted
		:type link: str
		:param manga: numbered id corresponding to the manga on the website, ex: 286
		:type manga: str
		:param chapstart: number of the chapter that the download is supposed to start
		:type chapstart: int/float/str

		:raises MangaNotFound: the scan asked for can't be found
		:raises EmptyChapter: the chapter you want to download is empty
		:raises DelayedRelease: the chapter you want to download isn't available yet
		"""

		super().__init__(link, manga, chapstart)

		if lang:
			self._lang = lang
		self.domain = ".mangadex.org"

		self.scrapper = cfscrape.create_scraper()

		# getting the manga id
		manga = str(manga)
		if link is not None:
			if link[-1] == "/":
				link = link[:-1]
			manga_id = link.split("/")[-2]
		elif manga.isdecimal():
			manga_id = manga
		else:
			# TODO: implement a search function for non id requests
			raise MangaNotFound(manga)


		try:
			manga_json = self.scrapper.get(f"https://mangadex.org/api/manga/{manga_id}").json()
		except json.decoder.JSONDecodeError:
			raise DownedSite("Mangadex")
		if manga_json.get("status") == "Manga ID does not exist.":
			raise MangaNotFound(manga_id)

		self._chapters_json = manga_json.get("chapter")

		self._ordered_chaps_json = []
		for item in sorted(self._chapters_json.keys(), key=lambda x: self._chapters_json[x]["chapter"].zfill(10)):
			self._ordered_chaps_json.append((item, self._chapters_json.get(item)))
		self._ordered_chaps_json = [chap for chap in self._ordered_chaps_json if chap[1].get("lang_code")==self._lang]

		self.manga_name = manga_json.get("manga").get("title")
		self.author = manga_json.get("manga").get("author")

		self.go_to_chapter(chapstart)

	def _set_current_chap_info(self, chap_id):
		"""
		Method used to update the internal info of the parser about the chapter given with the chap_id.

		:param chap_id: id of the chapter we want the info about

		:raises EmptyChapter: the chapter you want to download is empty
		:raises DelayedRelease: the chapter you want to download isn't available yet
		"""

		self.npage = 1


		try:
			self._current_chapter_json = self.scrapper.get(f"https://mangadex.org/api/chapter/{chap_id}").json()
		except json.decoder.JSONDecodeError:
			raise DownedSite("Mangadex")

		self.chapter_name = self._current_chapter_json.get("title").replace("/", "-")

		self.chapter_number = str(self._current_chapter_json.get("chapter")).split(".")[0].zfill(3)
		if "." in str(self._current_chapter_json.get("chapter")):
			self.chapter_number += "." + str(self._current_chapter_json.get("chapter")).split(".")[1]

		self._img_root = f"{self._current_chapter_json.get('server')}{self._current_chapter_json.get('hash')}/"
		self._img_ids = self._current_chapter_json.get("page_array")

		# checking if chapter is accessible
		if not self._img_ids:
			if self._current_chapter_json.get("status") == "delayed":
				ts = self._current_chapter_json.get("timestamp")
				date = datetime.fromtimestamp(ts)
				raise DelayedRelease(f"{self.manga_name}, chap {self.chapter_number}", date.isoformat(sep=" "))

			raise EmptyChapter(self.manga_name, self.chapter_number)

		self.image = f"{self._img_root}{self._img_ids[0]}"
		self.ext = self._img_ids[0].split(".")[1]

	def next_image(self):
		"""
		Goes to the next image in the scan being fetched.
		"""

		self.image = f"{self._img_root}{self._img_ids[self.npage]}"
		self.ext = self._img_ids[self.npage].split(".")[-1]
		self.npage += 1

	def is_last_image(self):
		"""
		Checks if it's the last image in the current chapter.
		:rtype: bool
		"""

		return self.npage == len(self._img_ids)

	def go_to_chapter(self, chap_num):
		"""
		Make the fetcher go to the asked chapter.

		:param chap_num: chapter number that was asked for
		:type chap_num: int/str/float

		:raises MangaNotFound: the asked chapter doesn't exist
		"""

		self._chap_id_pos = 0
		while self._chap_id_pos < len(self._ordered_chaps_json) and self._ordered_chaps_json[self._chap_id_pos][1].get("chapter") != str(chap_num):
			self._chap_id_pos += 1
		if self._chap_id_pos >= len(self._ordered_chaps_json):
			raise MangaNotFound(f"{self.manga_name}, chapter {chap_num}")

		self._set_current_chap_info(self._ordered_chaps_json[self._chap_id_pos][0])

	def next_chapter(self):
		"""
		Goes to the next chapter
		"""

		self._chap_id_pos += 1
		self._set_current_chap_info(self._ordered_chaps_json[self._chap_id_pos][0])

	def is_last_chapter(self):
		"""
		Checks if the current chapter is the last available one

		:rtype: bool
		"""

		return self._chap_id_pos+1 == len(self._ordered_chaps_json)

	@classmethod
	def scan(cls, link:str=None, manga:str=None, lang:str=""):

		header = {
			"User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/77.0.3865.120 Safari/537.36",
			"Set-Cookie": "domain=.mangadex.org",
			"Path": "/"
		}

		if lang:
			cls._lang = lang

		if link is not None:
			if link[-1] == "/":
				link = link[:-1]
			manga_id = link.split("/")[-2]
		elif manga.isdecimal():
			manga_id = manga
		else:
			raise MangaNotFound(manga)


		req = requests.get(f"https://mangadex.org/api/manga/{manga_id}", headers=header)
		if req.status_code == 200:
			manga_json = req.json()
		else:
			raise DownedSite("Mangadex")

		if manga_json.get("status") == "Manga ID does not exist.":
			raise MangaNotFound(manga_id)

		if manga_json.get("chapter"):
			chaps = [chap.get("chapter") for chap in manga_json.get("chapter").values() if chap.get("lang_code") == cls._lang and chap.get("chapter") != ""]
		else:
			return []

		return list({float(chap) if "." in chap else int(chap) for chap in chaps})

	def quit(self):
		"""
		Method used to close everything that was used after finishing to use the fetcher
		"""

		self.scrapper.close()


class MangadexEN(Mangadex):
	_lang = "gb"


class MangadexFR(Mangadex):
	_lang = "fr"




import os, sys
import json
import cfscrape
import re
from ..excepts import IsStandalone, FetcherNotFound, EmptyChapter, DelayedRelease
from ..Pyscandl import Pyscandl
from ..fetchers import FetcherEnum
from ..fetchers.fetcher import StandaloneFetcher


class Controller:
	__doc__ = """
	Object responsible of the autodl part of the program, all the logic related to it passes here
	"""

	def __init__(self, output:str=".", quiet:bool=False, tiny:bool=False):
		"""
		Initializes this instance of the autodl controller.
		If there is no json database for autodl currently in existence a new one is created at ``pyscandl/modules/autodl/db.json``.

		:param output: location where the outputted scans should be stored
		:type output: str
		:param quiet: should the program not output any information about what it is doing in the console
		:type quiet: bool
		:param tiny: should the name of every downloaded scan be minified and only include the chapter number and the chapter title
		:type tiny: bool
		"""

		try:
			with open(f"{os.path.dirname(sys.modules['pyscandl.modules.autodl'].__file__)}/db.json", "r") as data:
				self.db = json.load(data)
		except FileNotFoundError:
			self.db = {}
		self.output = output
		self.quiet = quiet
		self.tiny = tiny
		self._re_mgdex_scan = re.compile(r"(?:Chapter \d+, )?Chapter (?P<chap>\d+(\.\d+)?)")
		self.scrapper = cfscrape.create_scraper()
		self.missing_chaps = []
		self.downloads = 0

	def save(self):
		"""
		Saves the current state of the database in the ``db.json`` file.
		"""

		with open(f"{os.path.dirname(sys.modules['pyscandl.modules.autodl'].__file__)}/db.json", "w") as data:
			json.dump(self.db, data, indent=4, sort_keys=True)

	def add(self, name:str, link:str, fetcher:str, chapters:list=None, archived=False):
		"""
		Adds a new scan entry to the ``db.json`` file.

		:param name: name of the manga
		:type name: str
		:param link: link to the page of the manga *(same link that is used for the -l arg in other uses of pyscandl)*
		:type link: str
		:param fetcher: name of the associated fetcher
		:type fetcher: str
		:param chapters: list of the already possessed chapters that wont be downloaded again *(Optional)*
		:type chapters: list[int/float/str]
		:param archived: tell if the chapter is considered archived and if it'll be downloaded with autodl
		:type archived: bool

		:raises FetcherNotFound: the specified fetcher doesn't exist
		:raises IsStandalone: the specified fetcher is a standalone fetcher
		"""

		if chapters is None:
			chapters = []

		if fetcher.upper() not in [i.name for i in FetcherEnum]:
			raise FetcherNotFound(fetcher)
		if fetcher.lower() in ["nhentai"]:
			raise IsStandalone(name)
		self.db[name] = {
			"link": link,
			"fetcher": fetcher.upper(),
			"chapters": sorted(chapters, reverse=True),
			"archived": archived
		}

	def edit(self, name:str, link:str=None, fetcher=None, chapters:list=None, archived=None):
		"""
		Edits an already existing entry in the ``db.json`` file.
		The :param name: is mandatory to find the correct entry and every other parameter specified will overwrite the existing values.

		:param name: name of the manga
		:type name: str
		:param link: link to the page of the manga *(same link that is used for the -l arg in other uses of pyscandl)*
		:type link: str
		:param fetcher: name of the associated fetcher
		:type fetcher: str
		:param chapters: list of the already possessed chapters that wont be downloaded again
		:type chapters: list[int/float/str]
		:param archived: tell if the chapter is considered archived and if it'll be downloaded with autodl
		:type archived: bool

		:raises IsStandalone: the specified fetcher is a standalone fetcher
		"""

		if link is not None:
			self.db.get(name)["link"] = link
		if fetcher is not None:
			standalone_check = FetcherEnum.get(fetcher)
			if issubclass(standalone_check, StandaloneFetcher):
				raise IsStandalone(name)
			self.db.get(name)["fetcher"] = fetcher
		if chapters is not None:
			self.db.get(name)["chapters"] = sorted(self.db.get(name)["chapters"] + chapters, reverse=True)
		if archived is not None:
			try:
				self.db.get(name)["archived"] = archived
			except TypeError:
				print(f"\"{name}\" isn't a manga in the database, you may consider adding it with manga add")

	# each website/fetcher can have differently made xml from their rss so we need to treat them separately if need be
	def scan(self, name:str):
		"""
		Scans the asked manga for new and non downloaded chapters and adds them to the controller queue.

		:param name: name of the manga
		:type name: str
		"""

		self.missing_chaps.clear()
		manga = self.db.get(name)

		manga_fetcher = FetcherEnum.get(manga.get("fetcher"))
		manga_chapters = manga_fetcher.scan(link=manga.get("link"))

		self.missing_chaps = list(set(manga_chapters) - set(manga.get("chapters")))

		self.missing_chaps.sort()
		if not self.quiet:
			if self.missing_chaps:
				print(f"new chapter(s) for {name}: {', '.join(map(str, self.missing_chaps))}")
			else:
				print(f"no new chapter for {name}")

	def download(self, name:str, pdf:bool=True, keep:bool=False, image:bool=False):
		"""
		Start the download of the chapters of the asked manga that have their number in the scan results.

		:param name: name of the manga
		:type name: str
		:param pdf: tell if the result should be kept as a pdf
		:type pdf: bool
		:param keep: tell if the result should be kept as a pdf and as a collection of images
		:type keep: bool
		:param image: tell if the result should be kept as a collection of images
		:type image: bool
		"""

		manga = self.db.get(name)
		fetcher = FetcherEnum.get(manga.get("fetcher"))

		# initialize to the first downloadable chapter and download it
		ok = False
		for chapter_id in range(len(self.missing_chaps)):
			try:
				downloader = Pyscandl(fetcher, self.missing_chaps[chapter_id], self.output, link=manga.get("link"), pdf=pdf, image=image, keep=keep, quiet=self.quiet, tiny=self.tiny)

				bad_image = True
				while bad_image:  # protect against bad downloads
					try:
						if keep or image:
							downloader.keep_full_chapter()
						elif pdf:
							downloader.full_chapter()
						if not image:
							downloader.create_pdf()
						bad_image = False
					except IOError:
						print(f"problem during download, retrying {name} chapter {self.missing_chaps[chapter_id]}")
						downloader.go_to_chapter(self.missing_chaps[chapter_id])

				self.db.get(name).get("chapters").append(self.missing_chaps[chapter_id])
				self.downloads += 1

				self.missing_chaps = self.missing_chaps[chapter_id+1:]
				ok = True
				break
			except EmptyChapter:
				if not self.quiet:
					print(f"skipping {name} chapter {self.missing_chaps[chapter_id]}: empty, wont be added in the downloaded list")
			except DelayedRelease as e:
				if not self.quiet:
					print(e)

		# if chapters are left to doawnload proceeds with it
		if ok:
			for chapter_id in range(len(self.missing_chaps)):
				try:
					bad_image = True
					while bad_image:  # protect against bad downloads
						try:
							downloader.go_to_chapter(self.missing_chaps[chapter_id])

							if keep or image:
								downloader.keep_full_chapter()
							else:
								downloader.full_chapter()
							if not image:
								downloader.create_pdf()
							bad_image = False
						except IOError:
							print(f"problem during download, retrying {name} chapter {self.missing_chaps[chapter_id]}")

					self.db.get(name).get("chapters").append(self.missing_chaps[chapter_id])
					self.downloads += 1
				except EmptyChapter:
					if not self.quiet:
						print(f"skipping {name} chapter {self.missing_chaps[chapter_id]}: empty, wont be added in the downloaded list")
				except DelayedRelease as e:
					if not self.quiet:
						print(e)

			downloader.fetcher.quit()
			self.db.get(name).get("chapters").sort(reverse=True)

		# remove the directory if there is no chapter
		try:
			folders = list(os.walk(self.output))[1:]
			for folder in folders:
				if not folder[2]:
					os.rmdir(folder[0])
		except OSError:
			pass

	def list_mangas(self, all=False, only=False):
		"""
		Gives the list of all the names of the mangas in the ``db.json`` file. if the db is empty, returns None

		:param all: get also the archived mangas
		:type all: bool
		:param only: get only the archived mangas
		:type only: bool

		:rtype: list
		"""

		titles = []

		if not self.db:
			return None

		max_len = len(max(self.db.keys(), key=lambda x: len(x)))
		for title in self.db.keys():
			if self.db[title]["archived"] and (all or only):
				titles.append("{title:<{size}}|{}".format("    Archived", title=title, size=max_len + 4))
			elif not self.db[title]["archived"] and not only:
				titles.append("{title:<{size}}|".format(title=title, size=max_len + 4))
		return titles

	def manga_info(self, name):
		"""
		Fet the infos about a specific manga.

		:param name: name of the manga
		:type name: str

		:rtype: dict
		"""

		return self.db.get(name)

	def delete_manga(self, name):
		"""
		Deletes a manga from the ``db.json`` file.

		:param name: name of the manga

		:return: confirms the deletion
		:rtype: bool
		"""
		if name in self.db:
			del self.db[name]
			return True
		else:
			return False

	def rm_chaps(self, name, *rm_chaps):
		"""
		Remove the listed chapters from the asked manga

		:param name: name of he manga
		:type name: str
		:param rm_chaps: list of all the chapters that have to be removed
		:type rm_chaps: str

		:return: confirms the deletion
		:rtype: bool
		"""
		if name in self.db:
			self.db.get(name)["chapters"] = [chap for chap in self.db.get(name)["chapters"] if not chap not in rm_chaps]
			return True
		else:
			return False

	def db_import(self, path:str):
		"""
		Takes an external json file path and put its content as the new database for pyscandl.

		:param path: path to the .json file to import as the database for autodl and manga
		:type path: str
		"""

		with open(path, "r") as data:
			self.db = json.load(data)

	def db_export(self, path:str):
		"""
		Saves a copy of the current database to the file path specified.

		:param path: path to the save location, may be either a file or a folder, if it is a folder the filename will be db.json
		:type path: str

		:raises TypeError: if you specify the file name in the destination, the file extension must be .json
		"""

		if os.path.splitext(path)[1]:  # check if the path includes the filename
			if os.path.splitext(path)[1] != ".json":
				raise TypeError("the file must be a .json file !")

			path, filename = os.path.split(path)

		else:
			filename = "db.json"

		if not os.path.exists(path):
			os.makedirs(path)

		with open(os.path.join(path, filename), "w") as data:
			json.dump(self.db, data, indent=4, sort_keys=True)



from pyscandl.modules import arg_parser, Pyscandl
from pyscandl.modules.fetchers import FetcherEnum
from pyscandl.modules.autodl import Controller
from pyscandl.modules.excepts import DownedSite
from pyscandl import __version__
import xml.etree.ElementTree


def main():
	args = arg_parser.get_parser().parse_args()

	if args.version:
		print(f" The current version of pyscandl is {__version__}")

	elif args.subparser == "manual":
		fetcher = FetcherEnum.get(args.fetcher)
		pyscandl = Pyscandl(fetcher, chapstart=args.chapter_start, output=args.output, pdf=args.pdf,
									 keep=args.keep, image=args.image, all=args.all, link=args.link, manga=args.manga,
									 download_number=args.download_number, chapend=args.chapter_end, quiet=args.quiet,
									 skip=args.skip, tiny=args.tiny)
		pyscandl.full_download()

	elif args.subparser == "manga":

		if args.list or args.list_all or args.list_only:
			ml = Controller().list_mangas(all=args.list_all, only=args.list_only)
			if ml:
				list = "\n- " + "\n- ".join(ml)
				print(f"current mangas in the autodl db are:{list}")
			else:
				print("there are currently no mangas in autodl, you may consider adding some to it with manga add")

		elif args.import_db:
			json = Controller()
			json.db_import(args.import_db)
			json.save()

		elif args.export_db:
			Controller().db_export(args.export_db)

		elif args.manga_subparser == "info":
			infos = Controller().manga_info(args.name)
			if infos is None:
				print(f"manga '{args.name}' not in the list, you may consider adding it to it with manga add")
			elif infos.get("chapters"):
				print(f"{args.name}:\n",
					  f"\tmanga link: {infos.get('link')}\n",
					  f"\tfetcher: {infos.get('fetcher').upper()}\n",
					  f"\tnumber of chapters already downloaded: {len(infos.get('chapters'))}\n",
					  f"\tlast chapter downloaded: {infos.get('chapters')[0]}\n",
					  f"\tarchived: {infos.get('archived')}")
			else:
				print(f"{args.name}:\n",
					  f"\tmanga link: {infos.get('link')}\n",
					  f"\tfetcher: {infos.get('fetcher').upper()}\n",
					  f"\tno chapter downloaded yet\n",
					  f"\tarchived: {infos.get('archived')}")

		elif args.manga_subparser == "add":
			json = Controller()
			if args.chap:
				chaps = [float(chap) if "." in chap else int(chap) for chap in args.chap]
			else:
				chaps = []
			json.add(args.name, args.link, args.fetcher, chaps, args.archived)
			json.save()

		elif args.manga_subparser == "edit":
			json = Controller()
			if args.chap:
				chaps = [float(chap) if "." in chap else int(chap) for chap in args.chap]
			else:
				chaps = None

			if args.archive:
				archive = True
			elif args.unarchive:
				archive = False
			else:
				archive = None

			json.edit(args.name, args.link, args.fetcher, chaps, archive)
			json.save()

		elif args.manga_subparser == "chaplist":
			chaps = Controller().manga_info(args.name).get("chapters")
			print(f"the already downloaded chapters for {args.name} are: {' '.join([str(chap) for chap in chaps])}")

		elif args.manga_subparser == "rmchaps":
			json = Controller()
			if json.rm_chaps(args.name, args.chap):
				json.save()
				if not args.quiet:
					print(f"deletion of the chapters {', '.join(args.chap)} from {args.name} sucessfull")
			else:
				if not args.quiet:
					print(f"no chapters removed for {args.name}")

		elif args.manga_subparser == "delete":
			json = Controller()
			if json.delete_manga(args.name):
				json.save()
				if not args.quiet:
					print(f"deletion of {args.name} successful")
			else:
				if not args.quiet:
					print(f"manga {args.name} not found")

	elif args.subparser == "autodl":
		autodl = Controller(args.output, args.quiet, args.tiny)
		# to be sure to save progress done in case of interruption
		try:
			for name in autodl.db:
				# checking if the manga is archived
				if not autodl.db[name]["archived"]:
					# currently having problems with the xml tree fetching sometimes so giving a retry possibility to happen
					tries_left = 3
					while tries_left > 0:
						try:
							autodl.scan(name)
							success = True
							tries_left = 0
						except xml.etree.ElementTree.ParseError:
							if not args.quiet:
								print(f"problem with the fetching for {name}, retrying...")
							success = False
							tries_left -= 1
						except DownedSite:
							# the website can't be accessed for the time being so no retrying
							success = False
							tries_left = 0
					if success:
						try:
							autodl.download(name, pdf=args.pdf, keep=args.keep, image=args.image)
						except DownedSite:
							print(f"can't access {name}, please retry it later")
					elif not args.quiet:
						print(f"can't access {name}, please retry it later")
		except KeyboardInterrupt:
			if not args.quiet:
				print("\nmanual interruption")
		finally:
			autodl.save()
			if not args.quiet:
				print(f"{autodl.downloads} chapters downloaded")


if __name__ == "__main__":
	main()



import requests
from bs4 import BeautifulSoup
from .fetcher import Fetcher
from ..excepts import MangaNotFound


class Naver(Fetcher):
    def __init__(self, link:str=None, manga:str=None, chapstart=1, collection=""):
        super().__init__(link, manga, chapstart)
        if collection:
            self._collection = collection
        self.domain = ".comic.naver.com"

        if link is not None:
            self._manga_id = link.partition("titleId=")[-1]
        elif str(manga).isdigit():
            self._manga_id = manga
        else:
            raise MangaNotFound(manga)

        self._set_current_chap_info(self.chapter_number)

    def _set_current_chap_info(self, chap_id):
        self.npage = 1
        self.chapter_number = int(chap_id)
        self._chap_url = f"https://comic.naver.com/{self._collection}/detail.nhn?titleId={self._manga_id}&no={self.chapter_number}"
        self._chap_req = requests.get(self._chap_url)

        if self._chap_req.url == "https://comic.naver.com/main.nhn":
            raise MangaNotFound(self._manga_id)

        self._bs4 = BeautifulSoup(self._chap_req.text, "html.parser")

        if self._chap_req.url == f"https://comic.naver.com/{self._collection}/list.nhn?titleId={self._manga_id}" or int(self._chap_req.url.partition("&no=")[-1]) != int(self.chapter_number):
            raise MangaNotFound(f"{self._manga_id}, chapter {self.chapter_number}")

        self.author = self._bs4.find("div", class_="comicinfo").findChild("span", class_="wrt_nm").text.strip()
        self.chapter_name = self._bs4.find("div", class_="tit_area").findChild("h3").text
        self.manga_name = self._bs4.find("div", class_="comicinfo").findChild("div", class_="detail").findChild("h2").contents[0]

        self._img_list = self._bs4.find(class_="wt_viewer").findChildren("img")

        self.image = self._img_list[0].get("src")
        self.ext = self.image.split(".")[1]

    def next_image(self):
        self.image = self._img_list[self.npage].get("src")
        self.ext = self.image.split(".")[1]
        self.npage += 1

    def go_to_chapter(self, chap):
        self._set_current_chap_info(chap)

    def next_chapter(self):
        self._set_current_chap_info(self.chapter_number + 1)

    def is_last_image(self):
        return self.npage == len(self._img_list)

    def is_last_chapter(self):
        return self.chapter_number == int(self._bs4.find("div", class_="pg_area").findChild("span", class_="total").text)

    @classmethod
    def scan(cls, link:str=None, manga:str=None):
        if link is not None:
            manga_id = link.partition("titleId=")[-1]
        elif str(manga).isdigit():
            manga_id = manga
        else:
            raise MangaNotFound(manga)

        req = requests.get(f"https://comic.naver.com/webtoon/detail.nhn?titleId={manga_id}")

        if req.url == "https://comic.naver.com/main.nhn":
            raise MangaNotFound(manga_id)

        last_chap = int(req.url.split("&no=")[-1])

        return list(range(1, last_chap + 1))


class NaverWebtoon(Naver):
    _collection = "webtoon"  # set at class level for the scan method

    def __init__(self, link:str=None, manga:str=None, chapstart=1):
        super().__init__(link, manga, chapstart)


class NaverBestChallenge(Naver):
    _collection = "bestChallenge"  # set at class level for the scan method

    def __init__(self, link:str=None, manga:str=None, chapstart=1):
        super().__init__(link, manga, chapstart)


class NaverChallenge(Naver):
    _collection = "challenge"  # set at class level for the scan method

    def __init__(self, link:str=None, manga:str=None, chapstart=1):
        super().__init__(link, manga, chapstart)



from ..excepts import MangaNotFound, EmptyChapter
from .fetcher import Fetcher
import re
import requests
import secrets
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import ParseError


class Fanfox(Fetcher):
	__doc__ = """
	This is the fetcher in charge of https://fanfox.net/ 
	"""

	def __init__(self, link:str=None, manga:str=None, chapstart=1):
		"""
		Initializes the instance of the nhentai fetcher, it needs either manga or link to work.

		:param link: link of the scan wanted
		:type link: str
		:param manga: manga name with all the non alpha numeric characters with "_", ex: fullmetal_alchemist
		:type manga: str
		:param chapstart: number of the chapter that the download is supposed to start
		:type chapstart: int/float/str

		:raises MangaNotFound: the scan asked for can't be found
		"""

		super().__init__(link, manga, chapstart)
		self._header = {"User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/77.0.3865.120 Safari/537.36",
						"Referer": "test"}

		self.domain = ".fanfox.net"
		# creating the chapter link
		if link is not None:
			if link[-1] == "/":
				self._link = link
			else:
				self._link = link + "/"
			self.manga_name = self._link.split("/")[-2].replace("_", " ").title()
		else:
			self._link = f"https://fanfox.net/manga/{manga.replace(' ', '_').lower()}/"
			self.manga_name = manga.replace("_", " ").title().replace("/", "_")

		# regex used
		self._c_eval_args = re.compile(r"eval\(function\(p,a,c,k,e,d\).+?p;}\('(?P<pattern>.+?)',(?P<counter1>\d+?),(?P<counter2>\d+?),'(?P<str_table>.+?)'\.split")
		self._c_decode_word_replacer = re.compile(r"\b\w+\b")
		self._c_chap_name = re.compile(r"<p class=\"reader-header-title-2\">(?:Vol\.(?:\d+|TBD) )?Ch\.\d+(?:\.\d+)? ?(?P<chap_name>.*?)</p>")
		self._c_multi_max_page = re.compile(r"(?:data-page=\")(?P<page_num>\d+)")
		self._c_multi_cid = re.compile(r"(?:var chapterid =)(?P<cid>\d+)(?:;)")
		self._c_multi_page_list = re.compile(r"(?:var pix=\")(?P<link>.*?)\".*?\[(?P<chaps>\".*?\")\]")
		self._c_mono_page_list = re.compile(r"=\[(?P<chaps>'.*?')];")
		self._c_next_chap_num = re.compile(f"reader-main-next\" href=\"/manga/{self.manga_name.replace(' ', '_').lower()}" + r"/(?:v(?:TBD|\d+)/)?c(?P<next_chap>\d+(?:\.\d+)?)/1\.html\" title=.*?>Next Chapter</a>")

		req = requests.get(self._link)
		if req.url != self._link:
			raise MangaNotFound(self.manga_name)
		try:
			self.author = re.search(r"Author: <a href=.*? title=\"(.*?)\">", req.text).group(1)
		except AttributeError:
			self.author = "unknown"

		self.go_to_chapter(chapstart)

	def go_to_chapter(self, chap_num):
		"""
		Make the fetcher go to the asked chapter.

		:param chap_num: chapter number that was asked for
		:type chap_num: int/str/float

		:raises MangaNotFound: the asked chapter doesn't exist
		:raises EmptyChapter: the chapter exists but is void of images
		"""

		self._image_list = []
		self.npage = 1
		self.chapter_number = str(chap_num).split(".")[0].zfill(3)
		if "." in str(chap_num):
			self.chapter_number += "." + str(chap_num).split(".")[1]

		url = f"{self._link}c{self.chapter_number}/1.html"
		self._req = requests.get(url, cookies={"isAdult": "1"})
		if "<title>404</title>" in self._req.text:
			raise MangaNotFound(f"{self.manga_name}, chapter {self.chapter_number}")
		if '<p class="detail-block-content">No Images</p>' in self._req.text:
			raise EmptyChapter(self.manga_name, self.chapter_number)

		self._mono = self._req.text.count("dm5_key") == 1
		self.chapter_name = self._c_chap_name.search(self._req.text).group("chap_name").replace("/", "_")

		if self._mono:
			self._mono_go_to_chap()
		else:
			self._multi_go_to_chap()

		self.image = self._image_list[0]
		self.ext = self.image.split(".")[-1]

	def _mono_go_to_chap(self):
		"""
		Method used to get the obfuscated image info in mono type scan chapters.
		*mono type chapters are characterized by having all the images displayed one after another in a single page*
		"""

		out = self._decode(self._req.text)
		self._image_list = [f"https:{page.split('?')[0]}" for page in self._c_mono_page_list.search(out).group("chaps").replace("'", "").split(",")]
		self._last_page = len(self._image_list)

	def _multi_go_to_chap(self):
		"""
		Method used to get obfuscated image info in multi type scan chapters.
		*multi type chapters are characterized by having one image per page*
		"""

		cid = self._c_multi_cid.search(self._req.text).group("cid")
		self._last_page = max([int(i) for i in self._c_multi_max_page.findall(self._req.text)])

		for page in range(1, self._last_page+1):
			if page > len(self._image_list):
				req2 = requests.get(f"https://fanfox.net/chapterfun.ashx?cid={cid}&page={page}&key={secrets.token_hex(8)}", headers=self._header)
				out = self._decode(req2.text)
				root = self._c_multi_page_list.search(out).group("link")
				chaps = [i.split("?")[0] for i in self._c_multi_page_list.search(out).group("chaps").split('"') if i != "" and i != ","]

				for i in chaps:
					if i[0] != "/":
						i = "/" + i
					self._image_list.append(f"https:{root + i}")

	def _decode(self, obfuscated:str):
		"""
		Method used to de-obfuscate the web-pages of the scan to get the image info

		:param obfuscated: html code of the web-page of the scan
		:type obfuscated: str

		:return decoded: de-obfuscated image info that was contained in the page
		:rtype decoded: str
		"""

		match = self._c_eval_args.search(obfuscated)
		pattern = match.group("pattern").replace("\\", "")
		counter1 = int(match.group("counter1"))
		counter2 = int(match.group("counter2"))
		str_table = match.group("str_table").split("|")
		pattern_dict = {}

		def calc_dict_indices(iter_number):
			if iter_number < counter1:
				prefix = ""
			else:
				prefix = calc_dict_indices(int(iter_number / counter1))

			remainder = iter_number % counter1

			if remainder > 35:
				return prefix + chr(remainder + 29)
			else:
				return prefix + "0123456789abcdefghijklmnopqrstuvwxyz"[remainder]

		for i in range(counter2 - 1, -1, -1):
			dict_index = calc_dict_indices(i)
			pattern_dict[dict_index] = str_table[i] if str_table[i] else dict_index

		return self._c_decode_word_replacer.sub(repl=lambda word: pattern_dict[word.group()], string=pattern)

	def next_image(self):
		"""
		Goes to the next image in the scan being fetched.
		"""

		self.image = self._image_list[self.npage]
		self.npage += 1
		self.ext = self.image.split(".")[-1]

	def is_last_image(self):
		"""
		Checks if it's the last image in the current chapter.
		:rtype: bool
		"""

		# automatically ignoring the last page of add
		return self.npage == len(self._image_list)-1

	def next_chapter(self):
		"""
		Goes to the next chapter
		"""

		next = self._c_next_chap_num.search(self._req.text).group("next_chap")
		self.go_to_chapter(next)

	def is_last_chapter(self):
		"""
		Checks if the current chapter is the last available one

		:rtype: bool
		"""

		return not bool(self._c_next_chap_num.search(self._req.text))

	@classmethod
	def scan(cls, link:str=None, manga:str=None):
		if link is not None:
			manga_id = link.split("manga/")[1].split("/")[0]
		elif manga is not None:
			manga_id = manga.replace(" ", "_")

		rss_url = f"https://fanfox.net/rss/{manga_id}.xml"

		try:
			xml_root = ET.fromstring(requests.get(rss_url).text)
		except ParseError:
			raise MangaNotFound(manga_id)

		chap_list = []

		for item in xml_root.iter("item"):
			chap_nb = item.find("link").text.split("/")[-2][1:]
			if "." in chap_nb:
				chap_nb = float(chap_nb)
			else:
				chap_nb = int(chap_nb)

			chap_list.append(chap_nb)

		return list(set(chap_list))


	def quit(self):
		"""
		Method used to close everything that was used after finishing to use the fetcher
		"""
		pass



from ..excepts import MangaNotFound
from .fetcher import StandaloneFetcher
import requests


class NHentai(StandaloneFetcher):
	__doc__ = """
	This is the fetcher in charge of https://nhentai.net/
	The fetcher is of standalone type, it considers every manga as a unique non chaptered scan.
	"""

	def __init__(self, link:str=None, manga:int=None):
		"""
		Initializes the instance of the nhentai fetcher, it needs either manga or link to work.

		:param link: link of the scan wanted
		:type link: str
		:param manga: numbered id corresponding to the manga on the website, ex: 177013
		:type manga: str
		:param chapstart: number of the chapter that the download is supposed to start, it's not used here but needs to be in the definition to respect the fetcher api

		:raises MangaNotFound: the scan asked for can't be found
		"""

		super().__init__(link, manga)
		if link is None:
			self._manga_json = requests.get(f"https://nhentai.net/api/gallery/{manga}").json()
		else:
			if link[-1] == "/":
				manga = link.split("/")[-2]
			else:
				manga = link.split("/")[-1]
			self._manga_json = requests.get(f"https://nhentai.net/api/gallery/{manga}").json()
		# checking if manga exists
		if self._manga_json.get("error"):
			raise MangaNotFound(manga)

		self._corresponding_table = { 'j': "jpg", 'p': "png", 'g': "gif"}
		self.domain = ".nhentai.net"
		self.author = "unknown"

		# getting the author
		found = False
		i = 0
		while not found and i < len(self._manga_json.get("tags")):
			if self._manga_json.get("tags")[i].get("type") == "artist":
				found = True
				self.author = self._manga_json.get("tags")[i].get("name").title()
			else:
				i += 1

		self.npage = 1
		self._image_list = self._manga_json.get("images").get("pages")
		self.ext = self._corresponding_table.get(self._image_list[0].get('t'))
		self._image_root = f"https://i.nhentai.net/galleries/{self._manga_json.get('media_id')}/"
		self.image = f"{self._image_root}{self.npage}.{self.ext}"
		self.manga_name = "NSFW"
		self.chapter_number = 1
		self.chapter_name = f'{self._manga_json.get("title").get("pretty").replace("/", "-")} - {manga}'

	def next_image(self):
		"""
		Goes to the next image in the scan being fetched.
		"""

		self.ext = self._corresponding_table.get(self._image_list[self.npage].get('t'))
		self.npage += 1
		self.image = f"{self._image_root}{self.npage}.{self.ext}"

	def is_last_image(self):
		"""
		Checks if it's the last image in the current chapter
		:rtype: bool
		"""

		return self.npage == self._manga_json.get("num_pages")



from enum import Enum
from . import NHentai, Fanfox, MangadexEN, MangadexFR, NaverWebtoon, NaverBestChallenge, NaverChallenge, FRScan, WebtoonsEN, WebtoonsFR
from ..excepts import FetcherNotFound


class FetcherEnum(Enum):
	__doc__ = """
	Enumeration of all the fetchers available to the program at the moment
	"""

	NHENTAI = NHentai
	FANFOX = Fanfox
	MANGADEX = MangadexEN
	MANGADEX_FR = MangadexFR
	FRSCAN = FRScan
	NAVERWEBTOON = NaverWebtoon
	NAVERBESTCHALLENGE = NaverBestChallenge
	NAVERCHALLENGE = NaverChallenge
	WEBTOONS = WebtoonsEN
	WEBTOONS_FR = WebtoonsFR

	@classmethod
	def get(cls, fetcher_name_request: str):
		"""
		Class method used to  get the corresponding fetcher that was asked for

		:param fetcher_name_request: name of the asked fetcher
		:type fetcher_name_request: str

		:return: corresponding fetcher object

		:raises FetcherNotFound: the name given doesn't correspond to one of the available fetchers
		"""

		for i in cls:
			if i.name == fetcher_name_request.upper():
				return i.value
		raise FetcherNotFound(fetcher_name_request)

	@classmethod
	def list(cls):
		"""
		Gives a list of all the available fetchers at the moment.
		:rtype: list[str]
		"""
		return [i.name for i in cls]



import requests
from bs4 import BeautifulSoup

from pyscandl.modules.excepts import DryNoSauceHere, MangaNotFound
from pyscandl.modules.fetchers.fetcher import Fetcher


class Webtoons(Fetcher):
    def __init__(self, link: str = None, manga: str = None, chapstart=1, lang=""):
        super().__init__(chapstart=chapstart)

        if lang:
            self._lang = lang

        if manga is not None:
            self._manga_id = manga
        elif link is not None:
            self._manga_id = link.split("=")[1]
        else:
            raise DryNoSauceHere()

        self._cookies = {"ageGatePass": "true"}
        self.headers = {
            "referer": "http://www.webtoons.com/",
            "contentLanguage": self._lang
        }
        self.domain = ".webtoons.com"

        req = requests.get(f"https://webtoons.com/LANG/CATEGORY/NAME/list?title_no={self._manga_id}",
                           cookies=self._cookies, headers=self.headers)
        if req.status_code == 404:
            raise MangaNotFound(self._manga_id)
        soup = BeautifulSoup(req.text, features="html.parser")

        self.author = soup.find("meta", property="com-linewebtoon:webtoon:author").get("content")
        if not self.author:
            self.author = "TBD"

        self.manga_name = soup.find("meta", property="og:title").get("content")
        self._last_chap = int(soup.find(id="_listUl").findChild("li").get("data-episode-no"))

        self.go_to_chapter(chapstart)

    def next_image(self):
        self.image = self._imgs[self.npage]
        self.npage += 1
        self.ext = self.image.split(".")[-1]

    def is_last_image(self):
        return self.npage == len(self._imgs)

    def go_to_chapter(self, chapter_no):
        self.npage = 1
        self.chapter_number = chapter_no
        req = requests.get(
            f"https://www.webtoons.com/LANG/CATEGORY/NAME/CHAPTITLE/viewer?title_no={self._manga_id}&episode_no={self.chapter_number}",
            cookies=self._cookies,
            headers=self.headers
        )

        if req.status_code == 404:
            raise MangaNotFound(f"{self._manga_id}, chapter {chapter_no}")
        self._chap_soup = BeautifulSoup(req.text, features="html.parser")

        self.chapter_name = self._chap_soup.find("h1", class_="subj_episode").get("title")

        img_tags = self._chap_soup.find(id="_imageList").findChildren("img")
        self._imgs = [img.get("data-url").split("?")[0] for img in img_tags]
        self.image = self._imgs[self.npage - 1]
        self.ext = self.image.split(".")[-1]

    def next_chapter(self):
        self.go_to_chapter(self.chapter_number + 1)

    def is_last_chapter(self):
        return self.chapter_number == self._last_chap

    @classmethod
    def scan(cls, link: str = None, manga: str = None, lang: str = ""):
        if lang:
            cls._lang = lang

        if manga is not None:
            _manga_id = manga
        elif link is not None:
            _manga_id = link.split("=")[1]
        else:
            raise DryNoSauceHere()

        _cookies = {"ageGatePass": "true"}
        headers = {
            "referer": "http://www.webtoons.com/",
            "contentLanguage": cls._lang
        }

        req = requests.get(f"https://webtoons.com/LANG/CATEGORY/NAME/list?title_no={_manga_id}",
                           cookies=_cookies, headers=headers)
        if req.status_code == 404:
            raise MangaNotFound(_manga_id)
        soup = BeautifulSoup(req.text, features="html.parser")

        _last_chap = int(soup.find(id="_listUl").findChild("li").get("data-episode-no"))

        return [chap for chap in range(1, _last_chap + 1)]


class WebtoonsEN(Webtoons):
    _lang = "en"


class WebtoonsFR(Webtoons):
    _lang = "fr"



from setuptools import setup
from pyscandl import __version__

with open("README.rst", "r") as f:
    readme = f.read()

requires = [
    "requests~=2.24.0",
    "img2pdf~=0.3.6",
    "Pillow~=7.2.0",
    "cfscrape~=2.1.1",
    "beautifulsoup4~=4.9.1",
]

setup(
    name='pyscandl',
    version=__version__,
    packages=['pyscandl', 'pyscandl.modules', 'pyscandl.modules.autodl', 'pyscandl.modules.fetchers'],
    url='https://pyscandl.readthedocs.io/',
    license='BSD-3-Clause License',
    author='Thomas MONTERO | Ara0n',
    author_email='thomas99.montero@gmail.com',
    description='a scan downloader in python',
    keywords="scraping web scan download manga comics mangadex webtoon nhentai fanfox naver frscan english french",
    project_urls={
        "Documentation": "https://pyscandl.readthedocs.io",
        "Source Code": "https://github.com/Ara0n/pyscandl"
    },
    python_requires=">=3.7",
    entry_points={"console_scripts": ["pyscandl = pyscandl.main:main"]},
    long_description=readme,
    long_description_content_type='text/x-rst',
    install_requires=requires,
    package_data={
        "pyscandl": ["banlist/*"],
        "": ["README.rst"]
    }
)



from abc import ABC, abstractmethod


class Fetcher(ABC):
    __doc__ = """
    Abstract class defining how a fetcher is made and what are the mandatory components in it to be able to interact with pyscandl.
    
    The objecive of a fetcher is to gather the information specific of the manga asked in a way that it can be easily used for the creation of the downloaded scan in a pdf and/or image form.
    """

    @abstractmethod
    def __init__(self, link:str=None, manga:str=None, chapstart=1):
        """
        Initialization of the fetcher. for your fetcher to be able to work you'll need at either link or manga to be specified.

        :param link: link to the main page of the wanted scan
        :type link: str
        :param manga: unique id corresponding to the manga on the website that can be used instead of the link to access the manga
        :type manga: str
        :param chapstart: number of the first chapter of the download
        """

        self.author:str = "TBD"  #: name of the author of the manga default to TBD if none is found
        self.chapter_name:str = ""  #: name of the chapter currently in the fetcher defaults to an empty string if no name is found
        self.chapter_number = chapstart  #: number of the chapter currently in the fetcher
        self.domain:str = None  #: domain of the website the fetcher is currently on looks like: .domname.ext
        self.ext:str = None  #: extention of the image currently in the fetcher
        self.image:str = None  #: url to the image currently in the fetcher
        self.manga_name:str = None  #: name of the manga currentl in the fetcher
        self.npage:int = 1  #: number of the page the fetcher is currently on
        self.headers = {} #: if the fetching of the image requires specific parameters in the request header put them here

    @abstractmethod
    def next_image(self):
        """
        Go to the next image in the current chapter.
        """

        pass

    @abstractmethod
    def go_to_chapter(self, chap):
        """
        Go to the specified chapter.

        :param chap: number of the specified chapter

        :raises MangaNotFound: the asked chapter doesn't exist
        """

        pass

    @abstractmethod
    def next_chapter(self):
        """
        Go to the next chapter in order.
        """

        pass

    @abstractmethod
    def is_last_image(self):
        """
        Checks if it's the last image in the current chapter.

        :rtype: bool
        """

        pass

    @abstractmethod
    def is_last_chapter(self):
        """
        Checks if the current chapter is the last available one

        :rtype: bool
        """

        pass

    @classmethod
    @abstractmethod
    def scan(cls, link:str=None, manga:str=None):
        """
        It is a class method to avoid initializing a fetcher with a chapter for othing, which would take time and resources for nothing.
        As a result you onnly get either the manga link or the manga id to do the method.

        Gives a list of all the chapters available for that current manga.
        The number given must correspond to the number you give to specify the chapters you want to the fetcher.
        To facilitate the usage please return a list of int and floats for the *.x* chapters.

        :param link: link of the manga
        :type link: str
        :param manga: unique id of the manga
        :type manga: str

        :rtype: list[int/float]

        :raises MangaNotFound: the asked manga doesn't exist
        """

        pass

    def quit(self):
        """
        Closes everything used by the fetcher safely.
        """

        pass


class StandaloneFetcher(ABC):
    __doc__ = """
    Abstract class defining how a standalone type fetcher is made and what are the mandatory components in it to be able to interact with pyscandl.
    
    This type of fetcher is used for the website that do not reference their scans with mangas and chapters but as standalones, an example of a website doing that is https://nhentai.net.
    """

    @abstractmethod
    def __init__(self, link:str=None, manga:str=None):
        """
        Initialization of the fetcher. for your fetcher to be able to work you'll need at either link or manga to be specified.

        :param link: link to the main page of the wanted scan
        :type link: str
        :param manga: unique id corresponding to the manga on the website that can be used instead of the link to access the manga
        :type manga: str
        """

        self.author:str = "TBD"  #: name of the author of the manga default to TBD if none is found
        self.chapter_name:str = ""  #: name of the chapter currently in the fetcher defaults to an empty string if no name is found, for standalones if the manga_name attribute is used to convey a different info you can place the manga name here, see the NHentai fetcher for an example
        self.domain:str = None  #: domain of the website the fetcher is currently on looks like: .domname.ext
        self.ext:str = None  #: extention of the image currently in the fetcher
        self.image:str = None  #: url to the image currently in the fetcher
        self.manga_name:str = None  #: name of the manga currentl in the fetcher, can be used differently as this dictates the name of the folder they are stored in, for example, the fetcher NHentaiuse this to store everything in a NSFW folder and tags them as NSFW
        self.npage:int = 1  #: number of the page the fetcher is currently on
        self.header = {} #: if the fetching of the image requires specific parameters in the request header put them here

    @abstractmethod
    def next_image(self):
        """
        Go to the next image in the current chapter.
        """

        pass

    @abstractmethod
    def is_last_image(self):
        """
        Checks if it's the last image in the current chapter.

        :rtype: bool
        """

        pass

    def quit(self):
        """
        Closes everything used by the fetcher safely.
        """

        pass
